//d3di_video.c
#define INITGUID
#include "doomdef.h"
#include "d3dr_local.h"

//#define SOFTWARE_DEVICE

#ifdef SOFTWARE_DEVICE
#define D3DDEVICEIID IID_IDirect3DRGBDevice
#else
#define D3DDEVICEIID IID_IDirect3DHALDevice
#endif

#include "d3di_video.h"
#include "c_d3d.h"
#include "d3di_ddraw.h"
#include "i_gamma.h"
#include "i_overlay.h"
#include "tables.h"
#include "d3dv_video.h"
#include "d3dr_state.h"

typedef struct
{
    word	masks[3];
    int		shifts[3];
}pixelformat_t;

LPDIRECTDRAW4			pDD=NULL;
LPDIRECTDRAWSURFACE4	pDDSPrimary=NULL;
LPDIRECTDRAWSURFACE4	pDDSBack=NULL;
LPDIRECTDRAWSURFACE4	pDDSScreens[5];

LPDIRECT3D3				pD3D=NULL;
LPDIRECT3DDEVICE3		pD3DDevice=NULL;
LPDIRECTDRAWSURFACE4	pDDSZBuffer=NULL;
LPDIRECT3DVIEWPORT3		pD3DViewport=NULL;
LPDIRECTDRAWPALETTE		pDDPTexture=NULL;
DDPIXELFORMAT			TexturePixelFormat;
DDPIXELFORMAT			AlphaTexturePixelFormat;
DDPIXELFORMAT			ScreenPixelFormat;

PALETTEENTRY			CurPal[256];
word					CurPal16[256];
word					TexturePal[MAXPLAYERS][256];
word					AlphaTexturePal[MAXPLAYERS][256];
word					TextureAlphaMask;
word					TextureClearColor;
int						TextureAlphaBits=16;
dboolean				UseColorKey;
dboolean				BilinearFiltering;

int						ViewWindowX;
int						ViewWindowY;
int						ViewWidth;
int						ViewHeight;
static byte				*gtable;


void I_InitDD(void)
{
    LPDIRECTDRAW		pdd1;
    DDSURFACEDESC2		ddsd;
    HRESULT				hres;
    DDSCAPS2			ddscaps;
    int					i;
	LPDIRECTDRAWCLIPPER	pclipper;

    for (i=0;i<5;i++)
		pDDSScreens[i]=NULL;
    hres=DirectDrawCreate(NULL, &pdd1, NULL);
    if (hres!=DD_OK)
		I_Error("DirectDrawCreate Failed");
    hres=pdd1->lpVtbl->QueryInterface(pdd1, &IID_IDirectDraw4, (void **)&pDD);
    pdd1->lpVtbl->Release(pdd1);
    if (hres!=DD_OK)
		I_Error("QueryInterface(IDirectDraw4) Failed");
	if (InWindow)
	{
		hres=pDD->lpVtbl->SetCooperativeLevel(pDD, *D_API.pMainWnd, DDSCL_NORMAL);
		if (hres!=DD_OK)
			I_Error("DirectDraw::SetCooperativeLevel Failed");
	}
	else
	{
		hres=pDD->lpVtbl->SetCooperativeLevel(pDD, *D_API.pMainWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
		if (hres!=DD_OK)
			I_Error("DirectDraw::SetCooperativeLevel Failed");
		hres=pDD->lpVtbl->SetDisplayMode(pDD, ScreenWidth, ScreenHeight, 16, 0, 0);
		if (hres!=DD_OK)
			I_Error("SetDisplayMode(%d, %d) Failed", ScreenWidth, ScreenHeight);
		UpdateWindow(*D_API.pMainWnd);
    }
    ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
    ddsd.dwSize=sizeof(DDSURFACEDESC2);
	if (InWindow)
	{
	    ddsd.dwFlags=DDSD_CAPS;
		ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE;
	}
	else
	{
	    ddsd.dwFlags=DDSD_CAPS|DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE|DDSCAPS_3DDEVICE|DDSCAPS_COMPLEX|DDSCAPS_FLIP;
	    ddsd.dwBackBufferCount=1;
	}
    hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSPrimary, NULL);
    if (hres!=DD_OK)
		I_Error("CreateSurface(Primary) Failed");
	if (InWindow)
	{
	    ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
		ddsd.dwSize=sizeof(DDSURFACEDESC2);
	    ddsd.dwFlags=DDSD_CAPS|DDSD_WIDTH|DDSD_HEIGHT;
		ddsd.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_3DDEVICE;
		ddsd.dwWidth=ScreenWidth;
		ddsd.dwHeight=ScreenHeight;
		hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSBack, NULL);
		if (hres!=DD_OK)
			I_Error("CreateSurface(Buffer Screen) Failed");
		hres=pDD->lpVtbl->CreateClipper(pDD, 0, &pclipper, NULL);
		if (hres!=DD_OK)
			I_Error("CreateClipper Failed");
		pclipper->lpVtbl->SetHWnd(pclipper, 0, *D_API.pMainWnd);
		pDDSPrimary->lpVtbl->SetClipper(pDDSPrimary, pclipper);
		pclipper->lpVtbl->Release(pclipper);
	}
	else
	{
		ZeroMemory(&ddscaps, sizeof(DDSCAPS2));
		ddscaps.dwCaps=DDSCAPS_BACKBUFFER;
		hres=pDDSPrimary->lpVtbl->GetAttachedSurface(pDDSPrimary, &ddscaps, &pDDSBack);
		if (hres!=DD_OK)
			I_Error("GetAttachedSurface(BackBuffer) Failed");
	}
    pDDSScreens[0]=pDDSBack;
    ZeroMemory(&ScreenPixelFormat, sizeof(DDPIXELFORMAT));
    ScreenPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
    hres=pDDSBack->lpVtbl->GetPixelFormat(pDDSBack, &ScreenPixelFormat);
    if (hres!=DD_OK)
		I_Error("GetPixelFormat Failed");
	if (ScreenPixelFormat.dwRGBBitCount!=16)
		I_Error("Only works with 16-bit desktop");
    ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
    ddsd.dwSize=sizeof(DDSURFACEDESC2);
    ddsd.dwFlags=DDSD_CAPS|DDSD_WIDTH|DDSD_HEIGHT;
    ddsd.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;
    ddsd.dwWidth=ScreenWidth;
    ddsd.dwHeight=ScreenHeight;
    for (i=1;i<4;i++)
    {
		hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSScreens[i], NULL);
		if (hres!=DD_OK)
			I_Error("CreateSurface(Screen%d) Failed", i);
    }
    ddsd.dwHeight=ST_HEIGHT;
    hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSScreens[4], NULL);
    if (hres!=DD_OK)
		I_Error("CreateSurface(Screen4) Failed");
}

static int CountBits(DWORD i)
{
	DWORD	mask;
	int		count;

	count=0;
	for (mask=1;mask;mask<<=1)
	{
		if (mask&i)
			count++;
	}
	return(count);
}

static HRESULT CALLBACK TextureFormatEnum(LPDDPIXELFORMAT ddpf, dboolean *ptexture8bit)
{
	int		bits;
	//FUDGE 8 bit textures?
#if 0
    if (ddpf->dwFlags==(DDPF_RGB|DDPF_PALETTEINDEXED8))
    {
		*ptexture8bit=true;
		memcpy(&TexturePixelFormat, ddpf, sizeof(DDPIXELFORMAT));
		return(D3DENUMRET_CANCEL);
    }
#endif
    if ((ddpf->dwFlags==(DDPF_RGB|DDPF_ALPHAPIXELS))&&(ddpf->dwRGBBitCount==16))
	{
		bits=CountBits(ddpf->dwRGBAlphaBitMask);
		if (bits<TextureAlphaBits)
		{
			memcpy(&AlphaTexturePixelFormat, ddpf, sizeof(DDPIXELFORMAT));
			TextureAlphaBits=bits;
		}
	}
	else if ((ddpf->dwFlags==DDPF_RGB)&&(ddpf->dwRGBBitCount==16))
	{
		if (!TexturePixelFormat.dwFlags)
			memcpy(&TexturePixelFormat, ddpf, sizeof(DDPIXELFORMAT));
	}
    return(D3DENUMRET_OK);
}

void I_SetViewMatrix(float x, float y, float z, angle_t angle)
{
    D3DMATRIX	m;
    float		s;
    float		c;
	float		dx;
	float		dy;

    ZeroMemory(&m, sizeof(D3DMATRIX));

    s=F2D3D(finesine[(viewangle-angle)>>ANGLETOFINESHIFT]);
    c=F2D3D(finecosine[(viewangle-angle)>>ANGLETOFINESHIFT]);
    m._11=s;
    m._12=c;
    m._22=s;
    m._21=-c;
    //m._11=m._22=1;
    m._33=m._44=1;
	dx=x-fviewx;
	dy=y-fviewy;
	m._41=(dx+viewoffset)*viewsin-dy*viewcos;
	m._42=(dx-viewoffset)*viewcos+dy*viewsin;
    //m._41=dx;
    //m._42=dy;
    m._43=z-fviewz;
    pD3DDevice->lpVtbl->SetTransform(pD3DDevice, D3DTRANSFORMSTATE_WORLD, &m);
}

void I_InitTransforms(void)
{
    D3DMATRIX	m;
    ZeroMemory(&m, sizeof(D3DMATRIX));
    m._11 =  1.0f;
    m._23 =  1.0f;
    m._32 =  1.0f;
    m._44 =  1.0f;
    pD3DDevice->lpVtbl->SetTransform(pD3DDevice, D3DTRANSFORMSTATE_VIEW, &m);

    ZeroMemory(&m, sizeof(D3DMATRIX));
    m._11 =  1.0f;
    //m._22 =  1.0f;
    m._22 =  ScreenWidth/(float)ScreenHeight;
    m._34 =  1.0f;
    m._33 =  1.00f;
    m._43 = -1.00f;
    pD3DDevice->lpVtbl->SetTransform(pD3DDevice, D3DTRANSFORMSTATE_PROJECTION, &m);
}

static HRESULT CALLBACK EnumZBuffers(LPDDPIXELFORMAT pf, LPDDPIXELFORMAT dest)
{
    if (pf->dwFlags!=DDPF_ZBUFFER)
		return(D3DENUMRET_OK);
    memcpy(dest, pf, sizeof(DDPIXELFORMAT));
    if (pf->dwZBufferBitDepth==16)
		return(D3DENUMRET_CANCEL);
    else
		return(D3DENUMRET_OK);
}

void I_CreateZBuffer(void)
{
    HRESULT		hres;
    DDSURFACEDESC2	ddsd;

    ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
    ddsd.dwSize=sizeof(DDSURFACEDESC2);
    hres=pD3D->lpVtbl->EnumZBufferFormats(pD3D, &D3DDEVICEIID, (LPD3DENUMPIXELFORMATSCALLBACK)EnumZBuffers, &ddsd.ddpfPixelFormat);
    if (hres!=D3D_OK)
		I_Error("EnumZBufferFormats Failed");
    if (!ddsd.ddpfPixelFormat.dwFlags)
		I_Error("Z-Buffering not supported");
    ddsd.dwFlags=DDSD_CAPS|DDSD_PIXELFORMAT|DDSD_WIDTH|DDSD_HEIGHT;
    ddsd.dwHeight=ScreenHeight;
    ddsd.dwWidth=ScreenWidth;
#ifdef SOFTWARE_DEVICE
    ddsd.ddsCaps.dwCaps=DDSCAPS_ZBUFFER|DDSCAPS_SYSTEMMEMORY;
#else
    ddsd.ddsCaps.dwCaps=DDSCAPS_ZBUFFER|DDSCAPS_VIDEOMEMORY;
#endif
    hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSZBuffer, NULL);
    if (hres!=DD_OK)
		I_Error("CreateSurface(ZBuffer) Failed");
    hres=pDDSBack->lpVtbl->AddAttachedSurface(pDDSBack, pDDSZBuffer);
    if (hres!=DD_OK)
		I_Error("AddAttachedSurface(ZBuffer) Failed");
}

void I_SetRenderViewport(void)
{
    D3DVIEWPORT2	d3dvp;
	HRESULT			hres;

    ZeroMemory(&d3dvp, sizeof(D3DVIEWPORT2));
    d3dvp.dwSize=sizeof(D3DVIEWPORT2);
    d3dvp.dwX=ViewWindowX;
    d3dvp.dwY=ViewWindowY;
    d3dvp.dwWidth=ViewWidth;
    d3dvp.dwHeight=ViewHeight;
    d3dvp.dvClipX=-1;
    d3dvp.dvClipY=1;
//    d3dvp.dvClipY=ScreenHeight/(float)(ScreenWidth);
    d3dvp.dvClipWidth=2;
    d3dvp.dvClipHeight=2;
//    d3dvp.dvClipHeight=ScreenHeight*2/(float)ScreenWidth;
    d3dvp.dvMinZ=0;
    d3dvp.dvMaxZ=1;
    hres=pD3DViewport->lpVtbl->SetViewport2(pD3DViewport, &d3dvp);
    if (hres!=D3D_OK)
		I_Error("SetViewport2 Failed(%d,%d-%d,%d", d3dvp.dwX, d3dvp.dwY, d3dvp.dwWidth, d3dvp.dwHeight);
}

static void SetAlphaTest(D3DCMPFUNC cmpfunc, DWORD ref)
{
	HRESULT		hres;

	hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
	if (hres!=DD_OK)
		I_Error("Alpha testing not available");
	hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_ALPHAFUNC, cmpfunc);
	if (hres!=D3D_OK)
		I_Error("Unable to setup alpha testing");
	hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_ALPHAREF, ref);
	if (hres!=D3D_OK)
		I_Error("Unable to setup alpha testing");
}

void I_SetFilterMode(void)
{
	DWORD		mode;

	if (BilinearFiltering)
		mode=D3DTFG_LINEAR;
	else
		mode=D3DTFG_POINT;
    pD3DDevice->lpVtbl->SetTextureStageState(pD3DDevice, 0, D3DTSS_MAGFILTER, mode);
	if (BilinearFiltering)
		mode=D3DTFN_LINEAR;
	else
		mode=D3DTFN_POINT;
    pD3DDevice->lpVtbl->SetTextureStageState(pD3DDevice, 0, D3DTSS_MINFILTER, mode);
}

static word RGB2HiColor(byte *p, pixelformat_t *pixfmt)
{
    word	res;
    int		i;

    res=0;
    for (i=0;i<3;i++)
    {
		if (pixfmt->shifts[i]<0)
			res|=(gtable[p[i]]>>-pixfmt->shifts[i])&pixfmt->masks[i];
		else
			res|=(gtable[p[i]]<<pixfmt->shifts[i])&pixfmt->masks[i];
    }
    return(res);
}

int GetShiftBits(DWORD mask)
{
    int	n;

    n=0;
    while ((mask&1)==0)
    {
		mask>>=1;
		n++;
    }
    while (mask&1)
    {
		mask>>=1;
		n++;
    }
    return(n);
}

void I_ConvertPixelFormat(pixelformat_t *pixfmt, LPDDPIXELFORMAT ddpf)
{
    int		i;

    pixfmt->masks[0]=(word)ddpf->dwRBitMask;
    pixfmt->masks[1]=(word)ddpf->dwGBitMask;
    pixfmt->masks[2]=(word)ddpf->dwBBitMask;
    for (i=0;i<3;i++)
		pixfmt->shifts[i]=GetShiftBits(pixfmt->masks[i])-8;
}

void I_TranslatePalette2(word *dest, byte *palette, LPDDPIXELFORMAT ddpf, byte *trans)
{
    pixelformat_t	pixelformat;
	int				i;

    I_ConvertPixelFormat(&pixelformat, ddpf);
    for (i=0;i<256;i++)
	{
		dest[i]=(word)(RGB2HiColor(&palette[trans[i]*3], &pixelformat)|ddpf->dwRGBAlphaBitMask);
	}
}

void I_TranslatePalette(word *dest, byte *palette, LPDDPIXELFORMAT ddpf)
{
    pixelformat_t	pixelformat;
	word			*p;

    I_ConvertPixelFormat(&pixelformat, ddpf);
	p=dest;
    while (p<&dest[256])
	{
		*(p++)=(word)(RGB2HiColor(palette, &pixelformat)|ddpf->dwRGBAlphaBitMask);
		palette+=3;
	}
}

void I_SetTexturePalette(void)
{
	byte		*pal;
	int			i;
	byte		trans[3][256];

	pal=W_CacheLumpName ("PLAYPAL",PU_CACHE);
	gtable=gammatable[ManualGamma];
	I_TranslatePalette(TexturePal[0], pal, &TexturePixelFormat);
	I_TranslatePalette(AlphaTexturePal[0], pal, &AlphaTexturePixelFormat);
    for (i=0 ; i<256 ; i++)
    {
		if (i >= 0x70 && i<= 0x7f)
		{
			// map green ramp to gray, brown, red
			trans[0][i] = 0x60 + (i&0xf);
			trans[1][i] = 0x40 + (i&0xf);
			trans[2][i] = 0x20 + (i&0xf);
		}
		else
		{
			// Keep all other colors as is.
			trans[0][i] = trans[1][i] = trans[2][i] = i;
		}
    }
	for (i=1;i<MAXPLAYERS;i++)
	{
		I_TranslatePalette2(TexturePal[i], pal, &TexturePixelFormat, trans[i-1]);
		I_TranslatePalette2(AlphaTexturePal[i], pal, &AlphaTexturePixelFormat, trans[i-1]);
	}
}

void I_CheckMinTextureSize(void)
{
	int			mask;
	if (MinTextureSize<1)
		MinTextureSize=1;
	mask=0x40000000;
	while ((mask&MinTextureSize)==0)
		mask>>=1;
	MinTextureSize=mask;
}

void I_InitD3D(void)
{
    HRESULT			hres;
    dboolean		texture8bit;
//    PALETTEENTRY	pal[256];
	D3DDEVICEDESC	ddesc;
	D3DDEVICEDESC	heldesc;

    hres=pDD->lpVtbl->QueryInterface(pDD, &IID_IDirect3D3, (void **)&pD3D);
    if (hres!=DD_OK)
		I_Error("QueryInterface(IDirect3D3) Failed");
    I_CreateZBuffer();
    hres=pD3D->lpVtbl->CreateDevice(pD3D, &D3DDEVICEIID, pDDSBack, &pD3DDevice, NULL);
    if (hres!=D3D_OK)
		I_Error("Direct3D::CreateDevice Failed");
    hres=pD3D->lpVtbl->CreateViewport(pD3D, &pD3DViewport, NULL);
    if (hres!=D3D_OK)
		I_Error("CreateViewport Failed");
    hres=pD3DDevice->lpVtbl->AddViewport(pD3DDevice, pD3DViewport);
    if (hres!=DD_OK)
		I_Error("AddViewport Failed");
    hres=pD3DDevice->lpVtbl->SetCurrentViewport(pD3DDevice, pD3DViewport);
    if (hres!=D3D_OK)
		I_Error("SetCurentViewport Failed");
    ZeroMemory(&TexturePixelFormat, sizeof(DDPIXELFORMAT));
    ZeroMemory(&AlphaTexturePixelFormat, sizeof(DDPIXELFORMAT));
    texture8bit=false;
    hres=pD3DDevice->lpVtbl->EnumTextureFormats(pD3DDevice, (LPD3DENUMPIXELFORMATSCALLBACK)TextureFormatEnum, &texture8bit);
    if (hres!=DD_OK)
		I_Error("EnumTextureFormats Failed");
    if (TexturePixelFormat.dwFlags==0)
		I_Error("Unable to find suitable texture format");
	if (!UseColorKey)
	{
		if (AlphaTexturePixelFormat.dwFlags==0)
			I_Error("Unable to find suitable alpha texture format");
	}
//FUDGE:8bt
#if 0
    if (texture8bit)
    {
		hres=pDD->lpVtbl->CreatePalette(pDD, DDPCAPS_ALLOW256|DDPCAPS_8BIT, pal, &pDDPTexture, NULL);
		if (hres!=DD_OK)
			I_Error("CreatePalette Failed");
    }
    else
#endif
		pDDPTexture=NULL;
    I_InitTransforms();
    hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_ZENABLE, D3DZB_TRUE);
    if (hres!=D3D_OK)
		I_Error("ZBufferEnable Failed");
	I_SetFilterMode();
	if (UseColorKey)
	{
		hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_COLORKEYENABLE, TRUE);
		if (hres!=D3D_OK)
			I_Error("Color keying not suported");
		TextureClearColor=(word)(TexturePixelFormat.dwRBitMask|TexturePixelFormat.dwBBitMask);
	}
	else
	{
		hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
		if (hres!=D3D_OK)
			I_Error("Unable to enable alpha blending");
		hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA);
		if (hres!=D3D_OK)
			I_Error("Unable to set source blend factor");
		hres=pD3DDevice->lpVtbl->SetRenderState(pD3DDevice, D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA);
		if (hres!=D3D_OK)
			I_Error("Unable to set dest blend factor");
		TextureAlphaMask=(word)AlphaTexturePixelFormat.dwRGBAlphaBitMask;
	    TextureClearColor=0;
		ddesc.dwSize=sizeof(D3DDEVICEDESC);
		heldesc.dwSize=sizeof(D3DDEVICEDESC);
		hres=pD3DDevice->lpVtbl->GetCaps(pD3DDevice, &ddesc, &heldesc);
		if (hres!=D3D_OK)
			I_Error("D3DDevice::GetCaps Failed");

		if (ddesc.dpcTriCaps.dwAlphaCmpCaps&D3DPCMPCAPS_GREATER)
			SetAlphaTest(D3DCMP_GREATER, 0);
		else if (ddesc.dpcTriCaps.dwAlphaCmpCaps&D3DPCMPCAPS_GREATEREQUAL)
			SetAlphaTest(D3DCMP_GREATEREQUAL, 1);
		else if (ddesc.dpcTriCaps.dwAlphaCmpCaps&D3DPCMPCAPS_NOTEQUAL)
			SetAlphaTest(D3DCMP_NOTEQUAL, 0);
	}

	I_SetTexturePalette();

	I_CheckMinTextureSize();
	/*
	{
		D3DDEVICEDESC	dd;

		ZeroMemory(&dd, sizeof(dd));
		dd.dwSize=sizeof(dd);
		pD3DDevice->lpVtbl->GetCaps(pD3DDevice, &dd, NULL);
		if (dd.dwDevCaps&D3DDEVCAPS_DRAWPRIMTLVERTEX)
			I_Error("color keying is supported");
		else
			I_Error("color keying not supported");
	}
	*/
}

void I_InitGraphics(void)
{
    I_InitDD();
    I_InitD3D();
	I_SetGamma();
    I_InitOverlay();
}

void I_ShutdownGraphics(void)
{
    int		i;

    I_ShutdownOverlay();
    SAFE_RELEASE(pDDSZBuffer);
//FUDGE:8bt
//    SAFE_RELEASE(pDDPTexture);
    SAFE_RELEASE(pD3DViewport);
    SAFE_RELEASE(pD3DDevice);
    SAFE_RELEASE(pD3D);
    SAFE_RELEASE(pDDSPrimary);
    if (pDD)
    {
		for (i=1;i<5;i++)
			SAFE_RELEASE(pDDSScreens[i]);
		SAFE_RELEASE(pDD);
    }
}

void I_SetPalette(byte *palette)
{
//    PALETTEENTRY	*dest;
//    HRESULT			hres;
	int				i;

    gtable=gammatable[ManualGamma];
	I_TranslatePalette(CurPal16, palette, &ScreenPixelFormat);
#if 0
	dest=CurPal;
//FUDGE:8bt
    if (pDDPTexture)
	{
	    while (dest<&CurPal[256])
	    {
			dest->peRed=gtable[palette[0]];
			dest->peGreen=gtable[palette[1]];
			dest->peBlue=gtable[palette[2]];
			dest->peFlags=0;
			dest++;
			palette+=3;
		}
		hres=pDDPTexture->lpVtbl->SetEntries(pDDPTexture, 0, 0, 256, CurPal);
		if (hres!=DD_OK)
			I_Error("Palette::SetEntries Failed");
    }
#endif
	for (i=0;i<NUM_AM_COLORS;i++)
		RGB16AutoMapColors[i]=CurPal16[AutoMapColors[i]];
}

void I_WaitForFlip(void)
{
    HRESULT	hres;

    while (1)
    {
		hres=pDDSPrimary->lpVtbl->GetFlipStatus(pDDSPrimary, DDGFS_ISFLIPDONE);
		if (hres==DD_OK)
			return;
		if (hres!=DDERR_WASSTILLDRAWING)
			return;
    }
}

void I_StartDrawing(void)
{
	if (!InWindow)
		I_WaitForFlip();
}

void I_FinishUpdate(void)
{
//	HRESULT		hres;

//FUDGERETURN
	if (InWindow)
	{
		/*hres=*/
		pDDSPrimary->lpVtbl->Blt(pDDSPrimary, D_API.pWindowPos, pDDSBack, NULL, DDBLT_WAIT, NULL);
	}
	else
	{
		/*hres=*/
		pDDSPrimary->lpVtbl->Flip(pDDSPrimary, NULL, DDFLIP_WAIT);
	}
}

void I_ScreenShot(void)
{
}

void I_SetGamma(void)
{
    ManualGamma=*D_API.pGammaLevel;
	R_DumpTextures();
	I_SetTexturePalette();
	R_RecalcLightmaps();
}

