//DOOM3D
//Input Routines

#include <stdlib.h>
#include <dinput.h>

#include "doomstat.h"
#include "i_system.h"
#include "i_windoz.h"
#include "d_main.h"
#include "m_argv.h"

#define KEYBOARD_BUFFERSIZE 64

dboolean	GrabMouse;
dboolean	UseJoystick;
dboolean	UseMouse;
dboolean	DigiJoy;

LPDIRECTINPUT			pDI=NULL;
LPDIRECTINPUTDEVICE		pDIMouse=NULL;
LPDIRECTINPUTDEVICE		pDIKeyboard=NULL;
LPDIRECTINPUTDEVICE2	pDIJoystick=NULL;

const char qwerty[]="qwertyuiop";
const char asdf[]="asdfghjkl;'";
const char zxcv[]="\\zxcvbnm,./";

int TranslateKey(int key)
{
	int		rc;

	rc=key;
    switch(rc)
    {
	case DIK_LEFT:		rc = KEY_LEFTARROW;     break;
	case DIK_RIGHT:		rc = KEY_RIGHTARROW;    break;
	case DIK_DOWN:		rc = KEY_DOWNARROW;     break;
	case DIK_UP:		rc = KEY_UPARROW;       break;
	case DIK_ESCAPE:	rc = KEY_ESCAPE;        break;
	case DIK_NUMPADENTER:
	case DIK_RETURN:	rc = KEY_ENTER;         break;
	case DIK_TAB:		rc = KEY_TAB;           break;
	case DIK_F1:		rc = KEY_F1;            break;
	case DIK_F2:		rc = KEY_F2;            break;
	case DIK_F3:		rc = KEY_F3;            break;
	case DIK_F4:		rc = KEY_F4;            break;
	case DIK_F5:		rc = KEY_F5;            break;
	case DIK_F6:		rc = KEY_F6;            break;
	case DIK_F7:		rc = KEY_F7;            break;
	case DIK_F8:		rc = KEY_F8;            break;
	case DIK_F9:		rc = KEY_F9;            break;
	case DIK_F10:		rc = KEY_F10;           break;
	case DIK_F11:		rc = KEY_F11;           break;
	case DIK_F12:		rc = KEY_F12;           break;

	case DIK_BACK:		rc = KEY_BACKSPACE;     break;

	case DIK_GRAVE:		rc = KEY_PAUSE;         break;

	case DIK_NUMPADEQUALS:
	case DIK_EQUALS:	rc = KEY_EQUALS;        break;

	case DIK_SUBTRACT:
	case DIK_MINUS:		rc = KEY_MINUS;         break;

	case DIK_LSHIFT:
	case DIK_RSHIFT:	rc = KEY_RSHIFT;		break;

	case DIK_LCONTROL:
	case DIK_RCONTROL:	rc = KEY_RCTRL;			break;

	case DIK_LALT:		rc = KEY_LALT;			break;
	case DIK_RALT:		rc = KEY_RALT;			break;
	case DIK_SPACE:		rc = ' ';				break;
	case DIK_0:			rc='0';					break;
	case DIK_CAPITAL:	rc=KEY_CAPS;			break;
	case DIK_INSERT:	rc=KEY_INS;				break;
	case DIK_DELETE:	rc=KEY_DEL;				break;
	case DIK_HOME:		rc=KEY_HOME;			break;
	case DIK_END:		rc=KEY_END;				break;
	case DIK_PRIOR:		rc=KEY_PGUP;			break;
	case DIK_NEXT:		rc=KEY_PGDOWN;			break;

	default:
		if ((rc>=DIK_1)&&(rc<=DIK_9))
			rc=((rc-DIK_1)%10)+'1';
		else if ((rc>=DIK_Q)&&(rc<=DIK_RBRACKET))
			rc=qwerty[rc-DIK_Q];
		else if ((rc>=DIK_A)&&(rc<=DIK_APOSTROPHE))
			rc=asdf[rc-DIK_A];
		else if ((rc>=DIK_BACKSLASH)&&(rc<=DIK_SLASH))
			rc=zxcv[rc-DIK_BACKSLASH];
		else
			rc=0;
	}
	return(rc);
}

void I_ReacquireInput(void)
{
    if (pDIMouse)
		pDIMouse->lpVtbl->Acquire(pDIMouse);
    if (pDIKeyboard)
		pDIKeyboard->lpVtbl->Acquire(pDIKeyboard);
}

void I_InitKeyboard(void)
{
    HRESULT	hres;
    DIPROPDWORD	dipdw;

    hres=pDI->lpVtbl->CreateDevice(pDI, &GUID_SysKeyboard, &pDIKeyboard, NULL);
    if (hres!=DI_OK)
		I_Error("CreateDevice(SysKeyboard) Failed");
    hres=pDIKeyboard->lpVtbl->SetCooperativeLevel(pDIKeyboard, hMainWnd, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
    if (hres!=DI_OK)
		I_Error("Keyboard::SetCooperativeLevel Failed");
    hres=pDIKeyboard->lpVtbl->SetDataFormat(pDIKeyboard, &c_dfDIKeyboard);
    if (hres!=DI_OK)
		I_Error("Keyboard::SetDataFormat Failed");
    dipdw.diph.dwSize=sizeof(DIPROPDWORD);
    dipdw.diph.dwHeaderSize=sizeof(DIPROPHEADER);
    dipdw.diph.dwObj=0;
    dipdw.diph.dwHow=DIPH_DEVICE;
    dipdw.dwData=KEYBOARD_BUFFERSIZE;
    hres=pDIKeyboard->lpVtbl->SetProperty(pDIKeyboard, DIPROP_BUFFERSIZE, &dipdw.diph);
    if (hres!=DI_OK)
		I_Error("Keyboard::SetProperty(BufferSize) Failed");
    pDIKeyboard->lpVtbl->Acquire(pDIKeyboard);
}

void I_InitMouse(void)
{
    HRESULT	hres;

    hres=pDI->lpVtbl->CreateDevice(pDI, &GUID_SysMouse, &pDIMouse, NULL);
    if (hres!=DI_OK)
	{
		I_Printf("I_InitMouse:CreateDevice(SysMouse) Failed - mouse input disabled");
		pDIMouse=NULL;
		return;
	}
    GrabMouse=false;
	if (!InWindow)
		hres=pDIMouse->lpVtbl->SetCooperativeLevel(pDIMouse, hMainWnd, DISCL_EXCLUSIVE|DISCL_FOREGROUND);
    if (InWindow||(hres!=DI_OK))
    {
		if (!M_CheckParm("-nograbmouse"))
			GrabMouse=true;
		hres=pDIMouse->lpVtbl->SetCooperativeLevel(pDIMouse, hMainWnd, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
		if (hres!=DI_OK)
		{
			I_Printf("I_InitMouse:SetCooperativeLevel Failed - mouse input disabled");
			SAFE_RELEASE(pDIMouse);
			return;
		}
    }
    hres=pDIMouse->lpVtbl->SetDataFormat(pDIMouse, &c_dfDIMouse);
    if (hres!=DI_OK)
	{
		I_Printf("I_InitMouse:SetDataFormat Failed - mouse input disabled");
		SAFE_RELEASE(pDIMouse);
		return;
	}
    pDIMouse->lpVtbl->Acquire(pDIMouse);
}

BOOL CALLBACK JoystickEnumProc(LPDIDEVICEINSTANCE ddi, LPVOID guid)
{
	memcpy(guid, &ddi->guidInstance, sizeof(GUID));
	UseJoystick=true;
	I_Printf("I_InitInput:Using joystick:%s\n", ddi->tszInstanceName);
	return(DIENUM_STOP);
}

int		JoyXHigh;
int		JoyXLow;
int		JoyYHigh;
int		JoyYLow;

void I_InitJoystick(void)
{
	HRESULT				hres;
	GUID				guid;
	LPDIRECTINPUTDEVICE	pdid;
	DIPROPRANGE			dipr;

	UseJoystick=false;
	hres=pDI->lpVtbl->EnumDevices(pDI, DIDEVTYPE_JOYSTICK, (LPDIENUMDEVICESCALLBACK)JoystickEnumProc, &guid, DIEDFL_ATTACHEDONLY);
	if ((hres!=DI_OK)||!UseJoystick)
	{
		I_Printf("I_InitInput:No joysticks found\n");
		return;
	}

    hres=pDI->lpVtbl->CreateDevice(pDI, &guid, &pdid, NULL);
    if (hres!=DI_OK)
		I_Error("CreateDevice(Joystick) Failed");
	hres=pdid->lpVtbl->QueryInterface(pdid, &IID_IDirectInputDevice2, (void **)&pDIJoystick);
	if (hres!=DI_OK)
		I_Error("Joystick::QueryInterface(IDirectInputDevice2) Failed");
	pdid->lpVtbl->Release(pdid);

	if (!InWindow)
		hres=pDIJoystick->lpVtbl->SetCooperativeLevel(pDIJoystick, hMainWnd, DISCL_EXCLUSIVE|DISCL_FOREGROUND);
    if (InWindow||(hres!=DI_OK))
    {
		hres=pDIJoystick->lpVtbl->SetCooperativeLevel(pDIJoystick, hMainWnd, DISCL_NONEXCLUSIVE|DISCL_FOREGROUND);
		if (hres!=DI_OK)
			I_Error("Joystick::SetCooperativeLevel Failed");
    }
    hres=pDIJoystick->lpVtbl->SetDataFormat(pDIJoystick, &c_dfDIJoystick);
    if (hres!=DI_OK)
		I_Error("Joystick::SetDataFormat Failed");

	dipr.diph.dwSize=sizeof(DIPROPRANGE);
	dipr.diph.dwHeaderSize=sizeof(DIPROPHEADER);
	dipr.diph.dwObj=DIJOFS_X;
	dipr.diph.dwHow=DIPH_BYOFFSET;
	hres=pDIJoystick->lpVtbl->GetProperty(pDIJoystick, DIPROP_RANGE, (LPDIPROPHEADER)&dipr);
	if (hres!=DI_OK)
		I_Error("GetProperty(DIPROP_RANGEX) Failed");
	if (DigiJoy)
	{
		JoyXHigh=dipr.lMin+(int)((dipr.lMax-dipr.lMin)*0.6);
		JoyXLow=dipr.lMin+(int)((dipr.lMax-dipr.lMin)*0.4);
	}
	else
	{
		JoyXLow=(dipr.lMin+dipr.lMax)/2;
		JoyXHigh=dipr.lMax-JoyXLow;
	}
	dipr.diph.dwSize=sizeof(DIPROPRANGE);
	dipr.diph.dwHeaderSize=sizeof(DIPROPHEADER);
	dipr.diph.dwObj=DIJOFS_X;
	dipr.diph.dwHow=DIPH_BYOFFSET;
	hres=pDIJoystick->lpVtbl->GetProperty(pDIJoystick, DIPROP_RANGE, (LPDIPROPHEADER)&dipr);
	if (hres!=DI_OK)
		I_Error("GetProperty(DIPROP_RANGEY) Failed");
	if (DigiJoy)
	{
		JoyYHigh=dipr.lMin+(int)((dipr.lMax-dipr.lMin)*0.6);
		JoyYLow=dipr.lMin+(int)((dipr.lMax-dipr.lMin)*0.4);
	}
	else
	{
		JoyYLow=(dipr.lMin+dipr.lMax)/2;
		JoyYHigh=dipr.lMax-JoyYLow;
	}

    pDIJoystick->lpVtbl->Acquire(pDIJoystick);
}

void I_InitInput(void)
{
    HRESULT	hres;

    hres=DirectInputCreate(hAppInst, DIRECTINPUT_VERSION, &pDI, NULL);
    if (hres!=DI_OK)
		I_Error("DirectInputCreate Failed");
	if (UseMouse)
		I_InitMouse();
    I_InitKeyboard();
	if (UseJoystick)
		I_InitJoystick();
}

void I_ShutdownInput(void)
{
	if (pDIJoystick)
	{
		pDIJoystick->lpVtbl->Unacquire(pDIJoystick);
		pDIJoystick->lpVtbl->Release(pDIJoystick);
		pDIJoystick=NULL;
	}
    if (pDIKeyboard)
    {
		pDIKeyboard->lpVtbl->Unacquire(pDIKeyboard);
		pDIKeyboard->lpVtbl->Release(pDIKeyboard);
		pDIKeyboard=NULL;
    }
    if (pDIMouse)
    {
		pDIMouse->lpVtbl->Unacquire(pDIMouse);
		pDIMouse->lpVtbl->Release(pDIMouse);
		pDIMouse=NULL;
    }
    SAFE_RELEASE(pDI);
}

void I_ProcessJoystick(void)
{
	DIJOYSTATE	dijs;
	int			i;
	HRESULT		hres;
	event_t		event;
	float		f;

	hres=pDIJoystick->lpVtbl->Poll(pDIJoystick);
	if (hres!=DI_OK)
	{
		I_Error("help");
		return;
	}

    hres=pDIJoystick->lpVtbl->GetDeviceState(pDIJoystick, sizeof(DIJOYSTATE), &dijs);
    if (hres==DIERR_INPUTLOST)
    {
		hres=pDIJoystick->lpVtbl->Acquire(pDIJoystick);
		if (hres!=DI_OK)
			return;
	    hres=pDIJoystick->lpVtbl->GetDeviceState(pDIJoystick, sizeof(DIJOYSTATE), &dijs);
    }
    if (hres!=DI_OK)
		return;

	event.type=ev_joystick;
	event.data1=0;
	for (i=0;i<JOYSTICK_BUTTONS;i++)
	{
		if (dijs.rgbButtons[i]&0x80)
			event.data1|=1<<i;
	}
	if (DigiJoy)
	{
		if (dijs.lX>JoyXHigh)
			event.data2=1;
		else if (dijs.lX<JoyXLow)
			event.data2=-1;
		else
			event.data2=0;
		if (dijs.lY>JoyYHigh)
			event.data3=1;
		else if (dijs.lY<JoyYLow)
			event.data3=-1;
		else
			event.data3=0;
	}
	else
	{
		f=(dijs.lX-JoyXLow)/(float)JoyXHigh;
		event.data2=*(int *)&f;
		f=(dijs.lY-JoyYLow)/(float)JoyYHigh;
		event.data3=*(int *)&f;
	}
	D_PostEvent(&event);
}

void I_ProcessMouse(void)
{
    event_t			event;
    HRESULT			hres;
    DIMOUSESTATE	dims;
    int				i;

    if (GrabMouse)
		SetCursorPos(ScreenWidth/2+WindowPos.left, ScreenHeight/2+WindowPos.top);

    hres=pDIMouse->lpVtbl->GetDeviceState(pDIMouse, sizeof(DIMOUSESTATE), &dims);
    if (hres==DIERR_INPUTLOST)
    {
		hres=pDIMouse->lpVtbl->Acquire(pDIMouse);
		if (hres!=DI_OK)
			return;
		hres=pDIMouse->lpVtbl->GetDeviceState(pDIMouse, sizeof(DIMOUSESTATE), &dims);
    }
    if (hres!=DI_OK)
		return;
    event.type=ev_mouse;
    event.data1=0;
    for (i=0;i<MOUSE_BUTTONS;i++)
    {
		if (dims.rgbButtons[i]&0x80)
			event.data1|=1<<i;
    }
    event.data2=dims.lX<<2;
    event.data3=-dims.lY<<2;
    D_PostEvent(&event);
}

void I_ProcessKeyboard(void)
{
    DWORD			num;
    HRESULT			hres;
    DIDEVICEOBJECTDATA		data[KEYBOARD_BUFFERSIZE];
    LPDIDEVICEOBJECTDATA	pdata;
    event_t			event;

    num=KEYBOARD_BUFFERSIZE;
    hres=pDIKeyboard->lpVtbl->GetDeviceData(pDIKeyboard, sizeof(DIDEVICEOBJECTDATA), data, &num, 0);
    if (hres==DIERR_INPUTLOST)
    {
		hres=pDIKeyboard->lpVtbl->Acquire(pDIKeyboard);
		if (hres!=DI_OK)
			return;
		num=KEYBOARD_BUFFERSIZE;
		hres=pDIKeyboard->lpVtbl->GetDeviceData(pDIKeyboard, sizeof(DIDEVICEOBJECTDATA), data, &num, 0);
    }
    if (hres!=DI_OK)
		return;
    for (pdata=data;pdata<&data[num];pdata++)
    {
		if (pdata->dwData&0x80)
			event.type=ev_keydown;
		else
			event.type=ev_keyup;
		event.data1=TranslateKey(pdata->dwOfs);
		D_PostEvent(&event);
    }
}

void I_ProcessInput(void)
{
	if (pDIJoystick)
		I_ProcessJoystick();
    if (pDIMouse)
		I_ProcessMouse();
    if (pDIKeyboard)
		I_ProcessKeyboard();
}
