#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>
#include "setup.h"
#include "i_reg.h"
#include "apiinfo.h"
#include "doomdef.h"

/*
#define false 0
#define true (!false)
*/
#define IDC_DYNAMIC			1000

#define IDC_MOUSEBUTTONS	1100
#define IDC_JOYBUTTONS		1200
#define IDC_MOUSE			101
#define IDC_JOYSTICK		102
#define IDC_ANALOGUEJOY		103

typedef struct
{
	char		*name;
	HMODULE		hmod;
	DoomAPI_t	DoomAPI;
	apiinfo_t	api;
	HGLOBAL		hTemplate;
}moduleinfo_t;

typedef struct
{
	char	*name;
	int		def;
}action_t;

typedef struct
{
	int		code;
	char	*name;
}keyinfo_t;

HINSTANCE	hAppInst=NULL;

char		CommandLine[1024];

POINT		Resolutions[]=
{
	{320, 200},
	{320, 240},
	{512, 384},
	{640, 480},
	{800, 600},
	{1024, 768},
	{1152, 864},
	{1280, 1024},
	{1600, 1200},
	{0, 0}
};

action_t	Actions[]=
{
	{"use", ' '},
	{"fire", KEY_RCTRL},
	{"up", KEY_UPARROW},
	{"down", KEY_DOWNARROW},
	{"left", KEY_LEFTARROW},
	{"right", KEY_RIGHTARROW},
	{"run", KEY_RSHIFT},
	{"strafe", KEY_RALT},
	{"strafeleft", ','},
	{"straferight", '.'},
	{"autorun", KEY_CAPS},
	{NULL, 0}
};

#define NUM_ACTIONS		((sizeof(Actions)/sizeof(action_t))-1)

#define MJBUTTON(m, j) (((byte)(m))|(((word)(byte)(j))<<8))
#define GETMBUTTON(x) ((char)((x)&0xff))
#define GETJBUTTON(x) ((char)((x)>>8))

action_t	ButtonActions[]=
{
	{"fire",	MJBUTTON(0, 0)},
	{"forward",	MJBUTTON(1, -2)},
	{"strafe",	MJBUTTON(2, 1)},
	{"use",		MJBUTTON(3, 3)},
	{"run",		MJBUTTON(-2, 2)},
	{NULL, 0}
};

#define NUM_BUTTONACTIONS	((sizeof(ButtonActions)/sizeof(action_t))-1)

int		MouseButtons[NUM_BUTTONACTIONS];
int		JoyButtons[NUM_BUTTONACTIONS];

keyinfo_t	Keys[]=
{
	{KEY_RIGHTARROW,	"Rt"},
	{KEY_LEFTARROW,		"Lft"},
	{KEY_UPARROW,		"Up"},
	{KEY_DOWNARROW,		"Dn"},
	{KEY_ESCAPE,		"Escape"},
	{KEY_ENTER,			"Enter"},
	{KEY_TAB,			"Tab"},
	{KEY_BACKSPACE,		"Backsp"},
	{KEY_PAUSE,			"Pause"},
	{KEY_RSHIFT,		"Shift"},
	{KEY_RALT,			"Alt"},
	{KEY_RCTRL,			"Ctrl"},
	{KEY_EQUALS,		"+"},
	{KEY_MINUS,			"-"},
	{KEY_ENTER,			"Enter"},
	{KEY_CAPS,			"Caps"},
	{KEY_INS,			"Ins"},
	{KEY_DEL,			"Del"},
	{KEY_HOME,			"Ho"},
	{KEY_END,			"End"},
	{KEY_PGUP,			"PU"},
	{KEY_PGDOWN,		"PD"},
	{';',				";"},
	{'\'',				"'"},
	{'#',				"#"},
	{'\\',				"\\"},
	{',',				","},
	{'.',				"."},
	{'/',				"/"},
	{'[',				"["},
	{']',				"]"},
	{'*',				"*"},
	{' ',				"Space"},
	{KEY_F1,			"F1"},
	{KEY_F2,			"F2"},
	{KEY_F3,			"F3"},
	{KEY_F4,			"F4"},
	{KEY_F5,			"F5"},
	{KEY_F6,			"F6"},
	{KEY_F7,			"F7"},
	{KEY_F8,			"F8"},
	{KEY_F9,			"F9"},
	{KEY_F10,			"F10"},
	{KEY_F11,			"F11"},
	{KEY_F12,			"F12"},
	{0,NULL}
};

int KeyLayout[]=
{
	-2,-1,'1','2','3','4','5','6','7','8','9','0',KEY_MINUS,KEY_EQUALS,-2,KEY_BACKSPACE,
	KEY_INS,KEY_HOME,KEY_PGUP, -1,'/','*',0,

	-2,KEY_TAB,'q','w','e','r','t','y','u','i','o','p','[',']',
	-2,-1,KEY_DEL,KEY_END,KEY_PGDOWN,'7','8','9',0,

	-2,KEY_CAPS,'a','s','d','f','g','h','j','k','l',';','\'','#',-2,KEY_ENTER,
	-3,-1,'4','5','6',0,

	-2,KEY_RSHIFT,'\\','z','x','c','v','b','n','m',',','.','/',-3,KEY_RSHIFT,
	-1,KEY_UPARROW,-1,'1','2','3',0,

	-2,KEY_RCTRL,-2,KEY_LALT,-8,' ',-2,KEY_RALT,-2,KEY_RCTRL,
	KEY_LEFTARROW,KEY_DOWNARROW,KEY_RIGHTARROW,-2,'0','.',0,0
};

#define KEY_LAYOUT_SIZE		(sizeof(KeyLayout)/sizeof(int))

int		UseJoystick;
int		UseMouse;
int		DigiJoy;

void Moan(char *fmt, ...);

void GetKeyName(char *buff, int key)
{
	keyinfo_t	*pkey;

	if (((key>='a')&&(key<='z'))||((key>='0')&&(key<='9')))
	{
		buff[0]=(char)toupper(key);
		buff[1]=0;
		return;
	}
	for (pkey=Keys;pkey->name;pkey++)
	{
		if (pkey->code==key)
		{
			strcpy(buff, pkey->name);
			return;
		}
	}
	strcpy(buff, "Unknown");
}

void LoadResolution(HWND hCombo, HKEY hkey)
{
	int		width;
	int		height;
	POINT	*p;
	char	buff[12];

	width=RegReadInt(hkey, "ScreenWidth", 320);
	height=RegReadInt(hkey, "ScreenHeight", 200);
	for (p=Resolutions;p->x;p++)
	{
		if ((p->x==width)&&(p->y==height))
		{
			SendMessage(hCombo, CB_SETCURSEL, p-Resolutions, 0);
			return;
		}
	}
	sprintf(buff, "%dx%d", width, height);
	SetWindowText(hCombo, buff);
}

void SelectDll(HWND hCombo, HKEY hkey)
{
	char			name[256];
	moduleinfo_t	*info;
	int				count;
	int				i;

	name[0]=0;
	RegReadString(hkey, "DllName", name, 256);
	if (strlen(name)<=4)
	{
		SendMessage(hCombo, CB_SETCURSEL, 0, 0);
		return;
	}
	name[strlen(name)-4]=0;
	count=SendMessage(hCombo, CB_GETCOUNT, 0, 0);
	for (i=0;i<count;i++)
	{
		info=(moduleinfo_t *)SendMessage(hCombo, CB_GETITEMDATA, i, 0);
		if (stricmp(info->name, name)==0)
		{
			SendMessage(hCombo, CB_SETCURSEL, i, 0);
			return;
		}
	}
	SendMessage(hCombo, CB_SETCURSEL, 0, 0);
}

void LoadControls(HKEY hkey)
{
	action_t	*action;
	char		buff[256];
	int			i;

	for (action=Actions;action->name;action++)
	{
		sprintf(buff, "key_%s", action->name);
		action->def=RegReadInt(hkey, buff, action->def);
	}
	for (action=ButtonActions, i=0;action->name;action++, i++)
	{
		sprintf(buff, "mouseb_%s", action->name);
		MouseButtons[i]=RegReadInt(hkey, buff, GETMBUTTON(action->def));
		sprintf(buff, "joyb_%s", action->name);
		JoyButtons[i]=RegReadInt(hkey, buff, GETJBUTTON(action->def));
	}
	UseMouse=RegReadInt(hkey, "use_mouse", true);
	UseJoystick=RegReadInt(hkey, "use_joystick", false);
	DigiJoy=RegReadInt(hkey, "DigiJoy", false);
}

void LoadSettings(HWND hWnd)
{
	HKEY	hkey;
	LONG	rc;

	rc=RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\FudgeFactor\\Doom3D", 0, KEY_QUERY_VALUE, &hkey);
	if (rc!=ERROR_SUCCESS)
		hkey=NULL;

	rc=RegReadInt(hkey, "InWindow", false);
	CheckDlgButton(hWnd, IDC_FULLSCREEN, rc?BST_UNCHECKED:BST_CHECKED);

	rc=RegReadInt(hkey, "FixSprites", true);
	CheckDlgButton(hWnd, IDC_FIXSPRITES, rc?BST_CHECKED:BST_UNCHECKED);

	rc=RegReadInt(hkey, "HighSound", false);
	CheckDlgButton(hWnd, IDC_HISOUND, rc?BST_CHECKED:BST_UNCHECKED);

	LoadControls(hkey);

	LoadResolution(GetDlgItem(hWnd, IDC_RESOLUTION), hkey);

	SelectDll(GetDlgItem(hWnd, IDC_DLLNAME), hkey);

	if (hkey)
		RegCloseKey(hkey);
}

void CleanupModule(moduleinfo_t *info)
{
	if (!info)
		return;
	if (info->name)
		free(info->name);
	if (info->api.FreeDoomAPI)
		info->api.FreeDoomAPI();
	if (info->hmod)
		FreeLibrary(info->hmod);
	if (info->hTemplate)
		GlobalFree(info->hTemplate);
	free(info);
}

void AddDll(HWND hCombo, char *name)
{
	moduleinfo_t	*info;
	int				i;

	info=(moduleinfo_t *)malloc(sizeof(moduleinfo_t));
	ZeroMemory(info, sizeof(moduleinfo_t));
	i=strlen(name);
	info->name=(char *)malloc(i-3);
	strncpy(info->name, name, i-4);
	info->name[i-4]=0;
	info->hmod=LoadLibrary(name);
	if (!info->hmod)
	{
		CleanupModule(info);
		return;
	}
	info->DoomAPI=(DoomAPI_t)GetProcAddress(info->hmod, "DoomAPI");
	if (!info->DoomAPI)
	{
		CleanupModule(info);
		return;
	}
	info->api.size=sizeof(apiinfo_t);
	info->api.version=API_INFO_VERSION;
	if (info->DoomAPI(&info->api, NULL)!=API_OK)
	{
		CleanupModule(info);
		return;
	}
	i=SendMessage(hCombo, CB_ADDSTRING, 0, (LPARAM)info->name);
	SendMessage(hCombo, CB_SETITEMDATA, i, (LPARAM)info);
}

void LoadDlls(HWND hCombo)
{
	WIN32_FIND_DATA	ffd;
	HANDLE			hfind;
	BOOL			more;

	hfind=FindFirstFile("*.dll", &ffd);
	if (hfind==INVALID_HANDLE_VALUE)
		return;

	more=true;
	while (more)
	{
		AddDll(hCombo, ffd.cFileName);
		if (!FindNextFile(hfind, &ffd))
			more=false;
	}
	FindClose(hfind);
}

HHOOK	hTTHook=NULL;
HWND	hToolTip=NULL;

LRESULT CALLBACK GetMsgProc(int nCode, WPARAM wParam, LPARAM lParam)
{
	LPMSG	pmsg;
	MSG		msg;

	pmsg=(LPMSG)lParam;
	if (nCode<0)
		return(CallNextHookEx(hTTHook, nCode, wParam, lParam));
	switch (pmsg->message)
	{
	case WM_MOUSEMOVE:
	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP:
		msg.lParam=pmsg->lParam;
		msg.wParam=pmsg->wParam;
		msg.message=pmsg->message;
		msg.hwnd=pmsg->hwnd;
		SendMessage(hToolTip, TTM_RELAYEVENT, 0, (LPARAM)&msg);
		break;
	}
	return(CallNextHookEx(hTTHook, nCode, wParam, lParam));
}

void AddToolTip(HWND hWnd, int ID, char *tip)
{
	TOOLINFO	ti;

	ti.cbSize=sizeof(TOOLINFO);
	ti.uFlags=TTF_IDISHWND|TTF_SUBCLASS;
	ti.hwnd=hWnd;
	ti.uId=(UINT)GetDlgItem(hWnd, ID);
	ti.hinst=NULL;
	ti.lpszText=tip;
	SendMessage(hToolTip, TTM_ADDTOOL, 0, (LPARAM)&ti);
}

void InitToolTip(HWND hWnd)
{
	hToolTip=CreateWindowEx(0, TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		hWnd, NULL, hAppInst, NULL);
	if (!hToolTip)
		return;

	AddToolTip(hWnd, IDC_RESOLUTION, "Screen Resolution or Window size");
	AddToolTip(hWnd, IDC_STATICN(IDC_RESOLUTION), "Screen Resolution or Window size");
	AddToolTip(hWnd, IDC_FULLSCREEN, "Fullscreen or Windowed");
	AddToolTip(hWnd, IDC_FIXSPRITES, "Sprites-in-a-PWAD fix");
	AddToolTip(hWnd, IDC_HISOUND, "16 or 8 bit sound");
	AddToolTip(hWnd, IDC_DLLNAME, "Rendering DLL");
	AddToolTip(hWnd, IDC_STATICN(IDC_DLLNAME), "Rendering DLL");
	AddToolTip(hWnd, IDC_ADVANCED, "Advanced rendering options");
	AddToolTip(hWnd, IDC_NETGAME, "Set network game options");
	AddToolTip(hWnd, IDC_PARAMETERS, "Additional command line parameters");
	AddToolTip(hWnd, IDC_STATICN(IDC_PARAMETERS), "Additional command line parameters");
	AddToolTip(hWnd, IDOK, "Save settings and run Doom3D");
	AddToolTip(hWnd, IDC_SAVE, "Save settings and exit");
	AddToolTip(hWnd, IDCANCEL, "Exit without saving settings");

	//hTTHook=SetWindowsHookEx(WH_GETMESSAGE, (HOOKPROC)GetMsgProc, NULL, GetCurrentThreadId());
}

void InitMainDlg(HWND hWnd)
{
	HWND	hctrl;
	char	buff[10];
	POINT	*p;

	hctrl=GetDlgItem(hWnd, IDC_RESOLUTION);
	for (p=Resolutions;p->x;p++)
	{
		sprintf(buff, "%dx%d", p->x, p->y);
		SendMessage(hctrl, CB_ADDSTRING, 0, (LPARAM)buff);
	}
	SendMessage(hctrl, CB_SETCURSEL, 0, 0);
	LoadDlls(GetDlgItem(hWnd, IDC_DLLNAME));
	LoadSettings(hWnd);
	InitToolTip(hWnd);
}

void CleanupMainDlg(HWND hWnd)
{
	HWND			hCombo;
	int				i;
	int				count;
	moduleinfo_t	*info;

	hCombo=GetDlgItem(hWnd, IDC_DLLNAME);
	count=SendMessage(hCombo, CB_GETCOUNT, 0, 0);
	for (i=0;i<count;i++)
	{
		info=(moduleinfo_t *)SendMessage(hCombo, CB_GETITEMDATA, i, 0);
		if (info)
		{
			CleanupModule(info);
			SendMessage(hCombo, CB_SETITEMDATA, i, (LPARAM)NULL);
		}
	}
}

void SaveResolution(HWND hCombo, HKEY hkey)
{
	int		i;
	int		width;
	int		height;
	char	buff[12];
	char	*s;

	i=SendMessage(hCombo, CB_GETCURSEL, 0, 0);
	if (i==CB_ERR)
	{
		GetWindowText(hCombo, buff, 12);
		width=strtol(buff, &s, 10);
		while (*s&&!isdigit(*s))
			s++;
		height=strtol(s, NULL, 10);
		if (!width&&height)
			return;
	}
	else
	{
		width=Resolutions[i].x;
		height=Resolutions[i].y;
	}
	RegWriteInt(hkey, "ScreenWidth", width);
	RegWriteInt(hkey, "ScreenHeight", height);
}

void SaveDllName(HWND hCombo, HKEY hkey)
{
	int				i;
	char			buff[256];
	moduleinfo_t	*info;

	i=SendMessage(hCombo, CB_GETCURSEL, 0, 0);
	info=(moduleinfo_t *)SendMessage(hCombo, CB_GETITEMDATA, i, 0);
	strcpy(buff, info->name);
	strcat(buff, ".dll");
	RegWriteString(hkey, "DllName", buff);
}

void SaveControls(HKEY hkey)
{
	action_t	*action;
	int			i;
	char		buff[256];

	for (action=Actions;action->name;action++)
	{
		sprintf(buff, "key_%s", action->name);
		RegWriteInt(hkey, buff, action->def);
	}
	for (action=ButtonActions, i=0;action->name;action++, i++)
	{
		if (MouseButtons[i]!=-2)
		{
			sprintf(buff, "mouseb_%s", action->name);
			RegWriteInt(hkey, buff, MouseButtons[i]);
		}
		if (JoyButtons[i]!=-2)
		{
			sprintf(buff, "joyb_%s", action->name);
			RegWriteInt(hkey, buff, JoyButtons[i]);
		}
	}
	RegWriteInt(hkey, "use_mouse", UseMouse);
	RegWriteInt(hkey, "use_joystick", UseJoystick);
	RegWriteInt(hkey, "DigiJoy", DigiJoy);
}

void SaveSettings(HWND hWnd)
{
	LONG	rc;
	DWORD	disp;
	HKEY	hkey;

	rc=RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\FudgeFactor\\Doom3D", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hkey, &disp);
	if (rc!=ERROR_SUCCESS)
		return;

	SaveResolution(GetDlgItem(hWnd, IDC_RESOLUTION), hkey);

	rc=IsDlgButtonChecked(hWnd, IDC_FULLSCREEN);
	RegWriteInt(hkey, "InWindow", rc!=BST_CHECKED);
	rc=IsDlgButtonChecked(hWnd, IDC_FIXSPRITES);
	RegWriteInt(hkey, "FixSprites", rc==BST_CHECKED);
	rc=IsDlgButtonChecked(hWnd, IDC_HISOUND);
	RegWriteInt(hkey, "HighSound", rc==BST_CHECKED);

	SaveDllName(GetDlgItem(hWnd, IDC_DLLNAME), hkey);

	SaveControls(hkey);

	RegCloseKey(hkey);
}

LPWORD AlignPtr(LPWORD p)
{
	return((LPWORD)((((DWORD)p)+3)&~3));
}

LPWORD WriteWCString(LPWORD p, char *s)
{
//FUDGE: remove 50 char limit?
	p+=MultiByteToWideChar(CP_ACP, 0, s, -1, (LPWSTR)p, 50);
	return(p);
}

LPWORD WriteDlgTemplate(LPDLGTEMPLATE pdt, char *name, int width)
{
	LPWORD				ptr;

	ptr=(LPWORD)(pdt+1);
	pdt->style=DS_MODALFRAME|DS_SETFONT|WS_VISIBLE|WS_POPUP|WS_CAPTION;
	pdt->x=0;
	pdt->y=0;
	pdt->cx=width;
	*(ptr++)=0;
	*(ptr++)=0;
	ptr=WriteWCString(ptr, name);
	*(ptr++)=8;
	ptr=WriteWCString(ptr, "MS Shell Dlg");
	ptr=AlignPtr(ptr);
	return(ptr);
}

LPWORD WriteDlgItem(LPWORD ptr, char *name, int x, int y, int width, int height, DWORD style, int id, int type)
{
	LPDLGITEMTEMPLATE	pdit;

	pdit=(LPDLGITEMTEMPLATE)ptr;
	ptr=(LPWORD)(pdit+1);

	pdit->style=style;
	pdit->x=x;
	pdit->y=y;
	pdit->cx=width;
	pdit->cy=height;
	pdit->id=id;
	*(ptr++)=0xFFFF;
	*(ptr++)=type;
	if (name)
		ptr=WriteWCString(ptr, name);
	else
		*(ptr++)=0;
	*(ptr++)=0;
	ptr=AlignPtr(ptr);
	return(ptr);
}

HGLOBAL MakeAdvancedDlgTemplate(default_t *defaults)
{
	LPDLGTEMPLATE		pdt;
	LPWORD				ptr;
	HGLOBAL				hmem;
	int					n;
	int					y;
	default_t			*def;
	DWORD				style;

//FUDGE:work out required size?
	hmem=GlobalAlloc(GMEM_ZEROINIT, 1024);
	if (!hmem)
		Moan("Unable to allocate memory");

	pdt=(LPDLGTEMPLATE)GlobalLock(hmem);
	ptr=WriteDlgTemplate(pdt, "Advanced Settings", 128);

	y=4;
	n=0;
	for (def=defaults;def->name;def++)
	{
		ptr=WriteDlgItem(ptr, def->name, 4, y+2, 88, 8, WS_CHILD|WS_VISIBLE, IDC_STATIC, 0x82);

		style=WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_BORDER|ES_AUTOHSCROLL;
		if (n==0)
			style|=WS_GROUP;
		if (!def->isstring)
			style|=ES_NUMBER;
		ptr=WriteDlgItem(ptr, NULL, 92, y, 32, 12, style, IDC_DYNAMIC+n, 0x81);

		y+=16;
		n++;
	}

	ptr=WriteDlgItem(ptr, "OK", 12, y, 50, 14, WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|BS_DEFPUSHBUTTON, IDOK, 0x80);
	/*ptr=*/WriteDlgItem(ptr, "Cancel", 66, y, 50, 14, WS_CHILD|WS_VISIBLE|WS_TABSTOP|BS_PUSHBUTTON, IDCANCEL, 0x80);

	pdt->cdit=n*2+2;
	pdt->cy=y+18;

	GlobalUnlock(hmem);
	return(hmem);
}

void LoadAdvancedSettings(HWND hWnd, moduleinfo_t *info)
{
	int			id;
	default_t	*def;
	HKEY		hkey;
	char		buff[256];
	LONG		rc;

	id=IDC_DYNAMIC;
	strcpy(buff, "Software\\FudgeFactor\\Doom3D\\");
	strcat(buff, info->name);
	strcat(buff, ".dll");
	rc=RegOpenKeyEx(HKEY_CURRENT_USER, buff, 0, KEY_QUERY_VALUE, &hkey);
	if (rc!=ERROR_SUCCESS)
		hkey=NULL;
	for (def=info->api.info->defaults;def->name;def++)
	{
		if (def->isstring)
		{
			strcpy(buff, (char *)def->defaultvalue);
			RegReadString(hkey, def->name, buff, 256);
			SetDlgItemText(hWnd, id, buff);
		}
		else
		{
			rc=RegReadInt(hkey, def->name, def->defaultvalue);
			SetDlgItemInt(hWnd, id, rc, TRUE);
		}
		id++;
	}
	RegCloseKey(hkey);
}

void SaveAdvancedSettings(HWND hWnd, moduleinfo_t *info)
{
	HKEY		hkey;
	LONG		rc;
	DWORD		disp;
	int			id;
	default_t	*def;
	char		buff[256];

	strcpy(buff, "Software\\FudgeFactor\\Doom3D\\");
	strcat(buff, info->name);
	strcat(buff, ".dll");
	rc=RegCreateKeyEx(HKEY_CURRENT_USER, buff, 0, NULL, REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &hkey, &disp);
	if (rc!=ERROR_SUCCESS)
		return;

	id=IDC_DYNAMIC;
	for (def=info->api.info->defaults;def->name;def++)
	{
		if (def->isstring)
		{
			GetDlgItemText(hWnd, id, buff, 256);
			RegWriteString(hkey, def->name, buff);
		}
		else
		{
			rc=GetDlgItemInt(hWnd, id, NULL, TRUE);
			RegWriteInt(hkey, def->name, rc);
		}
		id++;
	}
	RegCloseKey(hkey);
}

BOOL CALLBACK AdvancedDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetWindowLong(hWnd, DWL_USER, lParam);
		LoadAdvancedSettings(hWnd, (moduleinfo_t *)lParam);
		return(TRUE);
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
			SaveAdvancedSettings(hWnd, (moduleinfo_t *)GetWindowLong(hWnd, DWL_USER));
			EndDialog(hWnd, TRUE);
			return(TRUE);
		case IDCANCEL:
			EndDialog(hWnd, TRUE);
			return(TRUE);
		}
		break;
	}
	return(FALSE);
}

void DoAdvancedDlg(HWND hWnd)
{
	moduleinfo_t	*info;
	int				i;

	i=SendDlgItemMessage(hWnd, IDC_DLLNAME, CB_GETCURSEL, 0, 0);
	if (i==CB_ERR)
		return;
	info=(moduleinfo_t *)SendDlgItemMessage(hWnd, IDC_DLLNAME, CB_GETITEMDATA, i, 0);
	if (!info)
		return;

	if (!info->hTemplate)
		info->hTemplate=MakeAdvancedDlgTemplate(info->api.info->defaults);
	if (-1==DialogBoxIndirectParam(hAppInst, info->hTemplate, hWnd, (DLGPROC)AdvancedDlgProc, (LPARAM)info))
		Moan("Help!");
}

void DroneToggle(HWND hWnd)
{
	if (IsDlgButtonChecked(hWnd, IDC_DRONE)==BST_UNCHECKED)
	{
		CheckDlgButton(hWnd, IDC_LEFT, BST_UNCHECKED);
		CheckDlgButton(hWnd, IDC_RIGHT, BST_UNCHECKED);
		CheckDlgButton(hWnd, IDC_BACK, BST_UNCHECKED);
	}
}

void DroneTypeToggle(HWND hWnd, int id)
{
	if (IsDlgButtonChecked(hWnd, id)==BST_CHECKED)
	{
		CheckDlgButton(hWnd, IDC_DRONE, BST_CHECKED);
		if (id!=IDC_LEFT)
			CheckDlgButton(hWnd, IDC_LEFT, BST_UNCHECKED);
		if (id!=IDC_RIGHT)
			CheckDlgButton(hWnd, IDC_RIGHT, BST_UNCHECKED);
		if (id!=IDC_BACK)
			CheckDlgButton(hWnd, IDC_BACK, BST_UNCHECKED);
	}
}

void InitNetgameDlg(HWND hWnd)
{
	HKEY	hkey;
	LONG	rc;
	int		i;
	char	buff[256];
	char	name[7];

	rc=RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\FudgeFactor\\Doom3D\\Netgame", 0, KEY_QUERY_VALUE, &hkey);
	if (rc!=ERROR_SUCCESS)
		hkey=NULL;

	SendDlgItemMessage(hWnd, IDC_NUMPLAYERS, EM_SETLIMITTEXT, 1, 0);
	rc=RegReadInt(hkey, "NumPlayers", 2);
	SetDlgItemInt(hWnd, IDC_NUMPLAYERS, rc, FALSE);

	SendDlgItemMessage(hWnd, IDC_PLAYER, EM_SETLIMITTEXT, 1, 0);
	rc=RegReadInt(hkey, "Player", 1);
	SetDlgItemInt(hWnd, IDC_PLAYER, rc, FALSE);

	rc=RegReadInt(hkey, "Port", 0x666);
	SetDlgItemInt(hWnd, IDC_PORTNUM, rc, FALSE);

	for (i=0;i<NUM_NETREG_NODES;i++)
	{
		sprintf(name, "Node%d", i);
		buff[0]=0;
		RegReadString(hkey, name, buff, 256);
		SetDlgItemText(hWnd, IDC_NODE1+i, buff);
	}

	rc=RegReadInt(hkey, "Flags", false);
	if (rc&NDF_DRONE)
		CheckDlgButton(hWnd, IDC_DRONE, rc?BST_CHECKED:BST_UNCHECKED);
	if (rc&NDF_BACK)
		CheckDlgButton(hWnd, IDC_BACK, rc?BST_CHECKED:BST_UNCHECKED);
	if (rc&NDF_LEFT)
		CheckDlgButton(hWnd, IDC_LEFT, rc?BST_CHECKED:BST_UNCHECKED);
	if (rc&NDF_RIGHT)
		CheckDlgButton(hWnd, IDC_RIGHT, rc?BST_CHECKED:BST_UNCHECKED);

	if (rc&NDF_DEATH2)
		CheckDlgButton(hWnd, IDC_DEATH2, BST_CHECKED);
	else if (rc&NDF_DEATHMATCH)
		CheckDlgButton(hWnd, IDC_DEATH2, BST_CHECKED);
	else
		CheckDlgButton(hWnd, IDC_COOP, BST_CHECKED);

	if (hkey)
		RegCloseKey(hkey);
}

void SaveNetgameSettings(HWND hWnd)
{
	HKEY	hkey;
	LONG	rc;
	int		i;
	char	buff[256];
	char	name[7];

	rc=RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\FudgeFactor\\Doom3D\\Netgame", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_SET_VALUE, NULL, &hkey, (LPDWORD)&i);
	if (rc!=ERROR_SUCCESS)
		return;
	rc=GetDlgItemInt(hWnd, IDC_NUMPLAYERS, NULL, FALSE);
	RegWriteInt(hkey, "NumPlayers", rc);

	rc=GetDlgItemInt(hWnd, IDC_PLAYER, NULL, FALSE);
	RegWriteInt(hkey, "Player", rc);

	rc=GetDlgItemInt(hWnd, IDC_PORTNUM, NULL, FALSE);
	RegWriteInt(hkey, "Port", rc);

	for (i=0;i<NUM_NETREG_NODES;i++)
	{
		GetDlgItemText(hWnd, IDC_NODE1+i, buff, 256);
		sprintf(name, "Node%d", i);
		RegWriteString(hkey, name, buff);
	}

	i=0;
	if (IsDlgButtonChecked(hWnd, IDC_DRONE)==BST_CHECKED)
		i|=NDF_DRONE;
	else if (IsDlgButtonChecked(hWnd, IDC_BACK)==BST_CHECKED)
		i|=NDF_BACK;
	else if (IsDlgButtonChecked(hWnd, IDC_LEFT)==BST_CHECKED)
		i|=NDF_LEFT;
	else if (IsDlgButtonChecked(hWnd, IDC_RIGHT)==BST_CHECKED)
		i|=NDF_RIGHT;

	if (IsDlgButtonChecked(hWnd, IDC_DEATH2)==BST_CHECKED)
		i|=NDF_DEATH2;
	else if (IsDlgButtonChecked(hWnd, IDC_DEATHMATCH)==BST_CHECKED)
		i|=NDF_DEATHMATCH;

	RegWriteInt(hkey, "Flags", i);

	RegCloseKey(hkey);
}

BOOL CALLBACK NetgameDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		InitNetgameDlg(hWnd);
		return(TRUE);
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_DRONE:
			DroneToggle(hWnd);
			return(TRUE);
		case IDC_LEFT:
		case IDC_RIGHT:
		case IDC_BACK:
			DroneTypeToggle(hWnd, LOWORD(wParam));
			return(TRUE);
		case IDOK:
			SaveNetgameSettings(hWnd);
			EndDialog(hWnd, TRUE);
			return(TRUE);
		case IDCANCEL:
			EndDialog(hWnd, FALSE);
			return(TRUE);
		}
		break;
	}
	return(FALSE);
}

void SetupNetgame(HWND hWnd)
{
	if (IsDlgButtonChecked(hWnd, IDC_NETGAME)!=BST_CHECKED)
		return;
	DialogBox(hAppInst, MAKEINTRESOURCE(NETGAME_DLG), hWnd, (DLGPROC)NetgameDlgProc);
}

void InitControlsDlg(HWND hWnd)
{
	int		action;
	char	buff[256];

	for (action=0;Actions[action].name;action++)
	{
		GetKeyName(buff, Actions[action].def);
		SetDlgItemText(hWnd, IDC_DYNAMIC+1+action*2, buff);
	}
	CheckDlgButton(hWnd, IDC_JOYSTICK, UseJoystick?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_MOUSE, UseMouse?BST_CHECKED:BST_UNCHECKED);
	CheckDlgButton(hWnd, IDC_ANALOGUEJOY, DigiJoy?BST_UNCHECKED:BST_CHECKED);

	for (action=0;action<NUM_BUTTONACTIONS;action++)
	{
		SetDlgItemInt(hWnd, IDC_MOUSEBUTTONS+action, MouseButtons[action], true);
		SendDlgItemMessage(hWnd, IDC_MOUSEBUTTONS+action, EM_LIMITTEXT, 2, 0);

		SetDlgItemInt(hWnd, IDC_JOYBUTTONS+action, JoyButtons[action], true);
		SendDlgItemMessage(hWnd, IDC_JOYBUTTONS+action, EM_LIMITTEXT, 2, 0);
	}
}

BOOL CALLBACK ChangeKeyDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int		id;

	switch (uMsg)
	{
	case WM_INITDIALOG:
		//InitChangeKeyDlg(hWnd);
		return(TRUE);
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDCANCEL:
			EndDialog(hWnd, 0);
			break;
		default:
			id=LOWORD(wParam)-IDC_DYNAMIC;
			if ((id>0)&&(id<KEY_LAYOUT_SIZE))
				EndDialog(hWnd, KeyLayout[id]);
			break;
		}
		break;
	}
	return(FALSE);
}

HGLOBAL MakeKeyDlgTemplate(void)
{
	LPDLGTEMPLATE		pdt;
	LPWORD				ptr;
	HGLOBAL				hmem;
	int					n;
	int					y;
	int					x;
	int					key;
	int					width;
	DWORD				style;
	char				buff[256];
	int					nextwidth;

	hmem=GlobalAlloc(GMEM_ZEROINIT, 10240);
	if (!hmem)
		Moan("Unable to allocate memory");

	pdt=(LPDLGTEMPLATE)GlobalLock(hmem);
	ptr=WriteDlgTemplate(pdt, "Select Key", 1280);

	x=4;
	y=4;
	n=0;
	key=0;
	width=0;
	nextwidth=1;
	while (KeyLayout[key])
	{
		if (KeyLayout[key]<-1)
		{
			nextwidth=-KeyLayout[key];
		}
		else
		{
			if (KeyLayout[key]>-1)
			{
				GetKeyName(buff, KeyLayout[key]);
				style=WS_CHILD|WS_VISIBLE|WS_TABSTOP|BS_PUSHBUTTON;
				if (n==0)
					style|=WS_GROUP;
				ptr=WriteDlgItem(ptr, buff, x, y, 18*nextwidth-4, 14, style, IDC_DYNAMIC+key, 0x80);
				n++;
			}
			x+=18*nextwidth;
			nextwidth=1;
		}
		key++;
		if (!KeyLayout[key])
		{
			if (x>width)
				width=x;
			x=4;
			y+=18;
			key++;
		}
	}

//	ptr=WriteDlgItem(ptr, "OK", 12, y, 50, 14, WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|BS_DEFPUSHBUTTON, IDOK, 0x80);
//	ptr=WriteDlgItem(ptr, "Cancel", 66, y, 50, 14, WS_CHILD|WS_VISIBLE|WS_TABSTOP|BS_PUSHBUTTON, IDCANCEL, 0x80);

	pdt->cdit=n;
	pdt->cx=width;
	pdt->cy=y;

	GlobalUnlock(hmem);
	return(hmem);
}

void DoChangeKey(HWND hWnd, int a)
{
	action_t	*action;
	HGLOBAL		htemplate;
	int			key;
	char		buff[256];

	action=&Actions[a];
	htemplate=MakeKeyDlgTemplate();
	key=DialogBoxIndirect(hAppInst, htemplate, hWnd, (DLGPROC)ChangeKeyDlgProc);
	if (key)
	{
		action->def=key;
		GetKeyName(buff, key);
		SetDlgItemText(hWnd, IDC_DYNAMIC+1+a*2, buff);
	}
	GlobalFree(htemplate);
}

void EndControlsDlg(HWND hWnd)
{
	int		action;

	for (action=0;action<NUM_BUTTONACTIONS;action++)
	{
		if (MouseButtons[action]!=-2)
		{
			MouseButtons[action]=GetDlgItemInt(hWnd, IDC_MOUSEBUTTONS+action, NULL, true);
			if (MouseButtons[action]>=MOUSE_BUTTONS)
				MouseButtons[action]=-1;
		}
		if (JoyButtons[action]!=-2)
		{
			JoyButtons[action]=GetDlgItemInt(hWnd, IDC_JOYBUTTONS+action, NULL, true);
			if (JoyButtons[action]>=JOYSTICK_BUTTONS)
				JoyButtons[action]=-1;
		}
	}
	if (IsDlgButtonChecked(hWnd, IDC_MOUSE)==BST_CHECKED)
		UseMouse=true;
	else
		UseMouse=false;
	if (IsDlgButtonChecked(hWnd, IDC_JOYSTICK)==BST_CHECKED)
		UseJoystick=true;
	else
		UseJoystick=false;
	if (IsDlgButtonChecked(hWnd, IDC_ANALOGUEJOY)==BST_CHECKED)
		DigiJoy=false;
	else
		DigiJoy=true;
}

BOOL CALLBACK ControlsDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int		id;

	switch (uMsg)
	{
	case WM_INITDIALOG:
		InitControlsDlg(hWnd);
		return(TRUE);
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDOK:
		case IDCANCEL:
			EndControlsDlg(hWnd);
			EndDialog(hWnd, TRUE);
			break;
		default:
			id=LOWORD(wParam)-IDC_DYNAMIC;
			if ((id>=0)&&(id<NUM_ACTIONS*2)&&((id&1)==0)&&(HIWORD(wParam)==BN_CLICKED))
			{
				DoChangeKey(hWnd, id/2);
				return(TRUE);
			}
			break;
		}
	}
	return(FALSE);
}

HGLOBAL MakeControlsDlgTemplate(void)
{
	LPDLGTEMPLATE	pdt;
	LPWORD			ptr;
	HGLOBAL			hmem;
	int				n;
	int				y;
	int				height;
	int				action;
	DWORD			style;

//FUDGE:work out required size?
	hmem=GlobalAlloc(GMEM_ZEROINIT, 2048);
	if (!hmem)
		Moan("Unable to allocate memory");
	pdt=(LPDLGTEMPLATE)GlobalLock(hmem);

	ptr=WriteDlgTemplate(pdt, "Controls", 224);
	y=4;
	n=0;
	for (action=0;action<NUM_ACTIONS;action++)
	{
		style=WS_CHILD|WS_VISIBLE|WS_TABSTOP;
		if (n==1)
			style|=WS_GROUP;
		ptr=WriteDlgItem(ptr, Actions[action].name, 4, y, 50, 14, style, IDC_DYNAMIC+n, 0x80);
		n++;

		ptr=WriteDlgItem(ptr, NULL, 58, y+2, 66, 8, WS_CHILD|WS_VISIBLE, IDC_DYNAMIC+n, 0x82);
		n++;

		y+=18;
	}

	height=y;
	y=4;
	ptr=WriteDlgItem(ptr, "&Buttons:", 128, y, 32, 8, WS_CHILD|WS_VISIBLE, IDC_STATIC, 0x82);
	ptr=WriteDlgItem(ptr, "Mouse", 160, y, 32, 8, WS_CHILD|WS_VISIBLE, IDC_STATIC, 0x82);
	ptr=WriteDlgItem(ptr, "Joystick", 192, y, 32, 8, WS_CHILD|WS_VISIBLE, IDC_STATIC, 0x82);
	y+=12;
	for (action=0;action<NUM_BUTTONACTIONS;action++)
	{
		ptr=WriteDlgItem(ptr, ButtonActions[action].name, 128, y+2, 28, 8, WS_CHILD|WS_VISIBLE, IDC_STATIC, 0x82);
		if (MouseButtons[action]!=-2)
		{
			ptr=WriteDlgItem(ptr, "", 160, y, 28, 12, WS_CHILD|WS_VISIBLE|WS_BORDER|ES_NUMBER, IDC_MOUSEBUTTONS+action, 0x81);
			n++;
		}
		if (JoyButtons[action]!=-2)
		{
			ptr=WriteDlgItem(ptr, "", 192, y, 28, 12, WS_CHILD|WS_VISIBLE|WS_BORDER|ES_NUMBER, IDC_JOYBUTTONS+action, 0x81);
			n++;
		}
		y+=16;
	}
	ptr=WriteDlgItem(ptr, "Use &Mouse", 128, y, 120, 8, WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|BS_AUTOCHECKBOX, IDC_MOUSE, 0x80);
	y+=12;
	ptr=WriteDlgItem(ptr, "Use &Joystick", 128, y, 120, 8, WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|BS_AUTOCHECKBOX, IDC_JOYSTICK, 0x80);
	y+=12;
	ptr=WriteDlgItem(ptr, "&Analogue Joystick", 128, y, 120, 8, WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|BS_AUTOCHECKBOX, IDC_ANALOGUEJOY, 0x80);
	y+=12;
	if (y>height)
		height=y;
	/*ptr=*/WriteDlgItem(ptr, "&Close", 87, height, 50, 14, WS_CHILD|WS_VISIBLE|WS_TABSTOP|WS_GROUP|BS_DEFPUSHBUTTON, IDOK, 0x80);

	pdt->cdit=n+7+NUM_BUTTONACTIONS;
	pdt->cy=height+18;

	GlobalUnlock(hmem);
	return(hmem);
}

void DoControlsDlg(HWND hWnd)
{
	HGLOBAL	htemplate;

	htemplate=MakeControlsDlgTemplate();
	DialogBoxIndirect(hAppInst, htemplate, hWnd, (DLGPROC)ControlsDlgProc);
	GlobalFree(htemplate);
}

BOOL CALLBACK MainDlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_INITDIALOG:
		InitMainDlg(hWnd);
		return(TRUE);
	case WM_DESTROY:
		CleanupMainDlg(hWnd);
		return(TRUE);
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_ADVANCED:
			DoAdvancedDlg(hWnd);
			return(TRUE);
		case IDC_NETGAME:
			SetupNetgame(hWnd);
			return(TRUE);
		case IDC_CONTROLS:
			DoControlsDlg(hWnd);
			return(0);
		case IDOK:
			SaveSettings(hWnd);
			strcpy(CommandLine, "Doom3D.exe ");
			GetDlgItemText(hWnd, IDC_PARAMETERS, CommandLine+11, 1005);
			if (IsDlgButtonChecked(hWnd, IDC_NETGAME)==BST_CHECKED)
				strcat(CommandLine, " -netreg");
			EndDialog(hWnd, TRUE);
			return(TRUE);
		case IDC_SAVE:
			SaveSettings(hWnd);
			EndDialog(hWnd, FALSE);
		case IDCANCEL:
			EndDialog(hWnd, FALSE);
			return(TRUE);
		}
		break;
	}
	return(FALSE);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	PROCESS_INFORMATION	pi;
	STARTUPINFO			si;
	int					rc;

	hAppInst=hInstance;
	InitCommonControls();
	rc=DialogBox(hAppInst, MAKEINTRESOURCE(MAIN_DLG), NULL, (DLGPROC)MainDlgProc);
	if (!rc)
		return(0);

	ZeroMemory(&si, sizeof(STARTUPINFO));
	si.cb=sizeof(STARTUPINFO);
	if (CreateProcess(NULL, CommandLine, NULL, NULL,
		FALSE, CREATE_DEFAULT_ERROR_MODE|DETACHED_PROCESS,
		NULL, NULL, &si, &pi))
	{
		CloseHandle(pi.hProcess);
		CloseHandle(pi.hThread);
	}

	return(0);
}

void Moan(char *fmt, ...)
{
	va_list	va;
	char	buff[1024];

	va_start(va, fmt);
	vsprintf(buff, fmt, va);
	va_end(va);
	MessageBox(NULL, buff, "Error", MB_OK);
	exit(1);
}
