#define INITGUID
// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// $Id:$
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// $Log:$
//
// DESCRIPTION:
//	DOOM graphics stuff for DirectX
//
//-----------------------------------------------------------------------------
#include <stdlib.h>
#include <ddraw.h>

#include <stdarg.h>
//#include "doomstat.h"
//#include "i_system.h"
//#include "i_input.h"
#include "swi_video.h"
//#include "i_windoz.h"
#include "swv_video.h"
//#include "m_argv.h"
//#include "d_main.h"

#include "i_gamma.h"
#include "swr_main.h"
#include "i_overlay.h"
#include "swi_ddraw.h"

#include "doomdef.h"

#include "c_sw.h"

#include "sw16_defs.h"

LPDIRECTDRAW4		pDD=NULL;
LPDIRECTDRAWSURFACE4	pDDSPrimary=NULL;
LPDIRECTDRAWSURFACE4	pDDSBack=NULL;
pixel_t			*BufferScreens[2];
int			ScreenNum;
LPDIRECTDRAWSURFACE4	pDDSScreens[2]={NULL, NULL};

dboolean				InWindow;
dboolean				SurfaceLost=false;

pixel_t			CurrentPalette[256];
pixel_t			RGBFuzzMask;

void I_WaitVBL(int count)
{
    Sleep(count);
}

void I_ShutdownGraphics(void)
{
    I_ShutdownOverlay();
    SAFE_RELEASE(pDDSScreens[0]);
    SAFE_RELEASE(pDDSScreens[1]);
    SAFE_RELEASE(pDDSPrimary);
    SAFE_RELEASE(pDD);
}



void I_BltScreen(LPDDSURFACEDESC2 ddsd)
{
    pixel_t	*src;
    pixel_t	*dest;
    int		y;

    src=screens[0];
    dest=(pixel_t *)ddsd->lpSurface;

    for (y=0;y<ScreenHeight;y++)
    {
		memcpy(dest, src, PIX2BYTE(ScreenWidth));
		dest+=ddsd->lPitch;
		src+=ScreenWidth;
    }
}

void I_WaitForFlip(void)
{
    HRESULT	hres;

    while (1)
    {
		hres=pDDSPrimary->lpVtbl->GetFlipStatus(pDDSPrimary, DDGFS_ISFLIPDONE);
		if (hres==DD_OK)
			return;
		if (hres!=DDERR_WASSTILLDRAWING)
			return;
			//I_Error("GetFlipStatus Failed");
    }
}

DWORD Loose2RGBBits(DWORD mask)
{
    DWORD	i;

    i=1;
    while (i&&!(i&mask))
		i<<=1;
    return(mask&~(i|(i<<1)));
}

void I_InitScreenSurface(void)
{
    DDSURFACEDESC2	ddsd;
    HRESULT		hres;

    ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
    ddsd.dwSize=sizeof(DDSURFACEDESC2);
    ddsd.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
    hres=pDDSPrimary->lpVtbl->GetPixelFormat(pDDSPrimary, &ddsd.ddpfPixelFormat);
    if (hres!=DD_OK)
		I_Error("GetPixelFormat Failed");
	if (InWindow&&((ddsd.ddpfPixelFormat.dwFlags!=DDPF_RGB)||(ddsd.ddpfPixelFormat.dwRGBBitCount!=16)))
		I_Error("Windowed mode requires 16-bit color");
    RGBFuzzMask=(pixel_t)Loose2RGBBits(ddsd.ddpfPixelFormat.dwRBitMask);
    RGBFuzzMask|=Loose2RGBBits(ddsd.ddpfPixelFormat.dwGBitMask);
    RGBFuzzMask|=Loose2RGBBits(ddsd.ddpfPixelFormat.dwBBitMask);
    ddsd.dwFlags=DDSD_WIDTH|DDSD_HEIGHT|DDSD_LPSURFACE|DDSD_PITCH|DDSD_PIXELFORMAT|DDSD_CAPS;
    ddsd.dwWidth=ScreenWidth;
    ddsd.dwHeight=ScreenHeight;
    ddsd.lPitch=PIX2BYTE(ScreenWidth);
    ddsd.lpSurface=BufferScreens[0];
	/*
    ddsd.ddpfPixelFormat.dwFlags=DDPF_PALETTEINDEXED8|DDPF_RGB;
    ddsd.ddpfPixelFormat.dwRGBBitCount=8;
	*/
    ddsd.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;
    hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSScreens[0], NULL);
    if (hres!=DD_OK)
		I_Error("CreateSurface(Screen0) Failed", hres, DDERR_INCOMPATIBLEPRIMARY, DDERR_INVALIDOBJECT);
    ddsd.lpSurface=BufferScreens[1];
    hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSScreens[1], NULL);
    if (hres!=DD_OK)
		I_Error("CreateSurface(Screen1) Failed", hres, DDERR_INCOMPATIBLEPRIMARY, DDERR_INVALIDOBJECT);
}

void I_DoBltWindowed(void)
{
	HRESULT		hres;

	hres=pDDSPrimary->lpVtbl->Blt(pDDSPrimary, D_API.pWindowPos, pDDSScreens[ScreenNum], NULL, DDBLT_ASYNC|DDBLT_WAIT, NULL);
	if (hres==DDERR_SURFACELOST)
		SurfaceLost=true;
}

void I_DoBlt(void)
{
	RECT	r;
    HRESULT	hres;

    r.left=r.top=0;
    r.right=ScreenWidth;
    r.bottom=ScreenHeight;
	if (!InWindow)
		I_WaitForFlip();

	if (InWindow)
		I_DoBltWindowed();
	else
	{
		//hres=pDDSBack->lpVtbl->Blt(pDDSBack, NULL, pDDSScreens[ScreenNum], NULL, DDBLT_ASYNC|DDBLT_WAIT, NULL);
		hres=pDDSBack->lpVtbl->BltFast(pDDSBack, 0, 0, pDDSScreens[ScreenNum], &r, DDBLTFAST_WAIT);
		if (hres==DDERR_SURFACELOST)
			SurfaceLost=true;
	}
    ScreenNum=1-ScreenNum;
    screens[0]=BufferScreens[ScreenNum];
}

void I_StartDrawing(void)
{
	if (SurfaceLost&&(InWindow||!*D_API.pInBackground))
	{
		SurfaceLost=false;
		pDD->lpVtbl->RestoreAllSurfaces(pDD);
	}
	if (!InWindow)
		pDDSPrimary->lpVtbl->Flip(pDDSPrimary, NULL, DDFLIP_WAIT);
}

//
// I_FinishUpdate
//
void I_FinishUpdate (void)
{
    I_DoBlt();
}


//
// I_ReadScreen
//
void I_ReadScreen (pixel_t* scr)
{
    memcpy (scr, screens[0], PIX2BYTE(ScreenWidth*ScreenHeight));
}


//
// Palette stuff.
//

int GetShiftBits(DWORD mask)
{
    int	n;

    n=0;
    while ((mask&1)==0)
    {
		mask>>=1;
		n++;
    }
    while (mask&1)
    {
		mask>>=1;
		n++;
    }
    return(n);
}

int		shifts[3];
DWORD		masks[3];

word RGB2HiColor(byte *p)
{
    word	res;
    int		i;

    res=0;
    for (i=0;i<3;i++)
    {
		if (shifts[i]<0)
			res|=(p[i]>>-shifts[i])&masks[i];
		else
			res|=(p[i]<<shifts[i])&masks[i];
    }
    return(res);
}

//
// I_SetPalette
//
void I_SetPalette (byte* palette)
{
    word		*dest;
    byte		temp[3];
    DDPIXELFORMAT	ddpf;
    HRESULT		hres;
    int			i;
    int			map;
    byte		*table;

    ZeroMemory(&ddpf, sizeof(DDPIXELFORMAT));
    ddpf.dwSize=sizeof(DDPIXELFORMAT);
    hres=pDDSPrimary->lpVtbl->GetPixelFormat(pDDSPrimary, &ddpf);
    if (hres!=DD_OK)
		I_Error("GetPixelFormat Failed");
    masks[0]=ddpf.dwRBitMask;
    masks[1]=ddpf.dwGBitMask;
    masks[2]=ddpf.dwBBitMask;
    for (i=0;i<3;i++)
		shifts[i]=GetShiftBits(masks[i])-8;

    table=gammatable[ManualGamma];
    for (dest=CurrentPalette;dest<&CurrentPalette[256];dest++)
    {
		for (i=0;i<3;i++)
			temp[i]=table[palette[i]];
		*dest=RGB2HiColor(temp);
		for (map=0;map<NumColorMaps;map++)
		{
			for (i=0;i<3;i++)
				temp[i]=table[((int)palette[i])*(NumColorMaps+1-map)/(NumColorMaps+1)];
			colormaps[map][dest-CurrentPalette]=RGB2HiColor(temp);
		}
		temp[0]=temp[1]=temp[2]=table[255-(((int)palette[0]+(int)palette[1]+(int)palette[2])/3)];
		colormaps[NumColorMaps][dest-CurrentPalette]=RGB2HiColor(temp);
        palette+=3;
    }

	for (i=0;i<NUM_AM_COLORS;i++)
		RGB16AutoMapColors[i]=CurrentPalette[AutoMapColors[i]];
}

void I_InitGraphics(void)
{
    HRESULT		hres;
    LPDIRECTDRAW	dd;
    DDSURFACEDESC2	ddsd;
    DDSCAPS2		ddscaps;
	LPDIRECTDRAWCLIPPER	pclipper;

    if (pDD)
		return;

    hres=DirectDrawCreate(NULL, &dd, NULL);
    if (hres!=DD_OK)
		I_Error("DirectDrawCreate Failed");
    hres=dd->lpVtbl->QueryInterface(dd, &IID_IDirectDraw4, (void **)&pDD);
    if (hres!=DD_OK)
		I_Error("QueryInterface(IDirectDraw4) Failed");
    dd->lpVtbl->Release(dd);
	if (InWindow)
	{
		hres=pDD->lpVtbl->SetCooperativeLevel(pDD, *D_API.pMainWnd, DDSCL_NORMAL);
		if (hres!=DD_OK)
			I_Error("DirectDraw::SetCooperativeLevel Failed");
	}
	else
	{
		hres=pDD->lpVtbl->SetCooperativeLevel(pDD, *D_API.pMainWnd, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);
		if (hres!=DD_OK)
			I_Error("DirectDraw::SetCooperativeLevel Failed");
		hres=pDD->lpVtbl->SetDisplayMode(pDD, ScreenWidth, ScreenHeight, 16, 0, 0);
		if (hres!=DD_OK)
			I_Error("SetDisplayMode(%d, %d) Failed", ScreenWidth, ScreenHeight);
		UpdateWindow(*D_API.pMainWnd);
	}
    ZeroMemory(&ddsd, sizeof(DDSURFACEDESC2));
    ddsd.dwSize=sizeof(DDSURFACEDESC2);
	if (InWindow)
	{
	    ddsd.dwFlags=DDSD_CAPS;
		ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE;
	}
	else
	{
	    ddsd.dwFlags=DDSD_CAPS|DDSD_BACKBUFFERCOUNT;
		ddsd.dwBackBufferCount=1;
		ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE|DDSCAPS_COMPLEX|DDSCAPS_FLIP;
	}
    hres=pDD->lpVtbl->CreateSurface(pDD, &ddsd, &pDDSPrimary, NULL);
    if (hres!=DD_OK)
		I_Error("Unable to create primary surface");
    ScreenNum=0;
    BufferScreens[0]=screens[0];
    BufferScreens[1]=(pixel_t *)malloc(PIX2BYTE(ScreenWidth*ScreenHeight));
	if (InWindow)
	{
		hres=pDD->lpVtbl->CreateClipper(pDD, 0, &pclipper, NULL);
		if (hres!=DD_OK)
			I_Error("CreateClipper Failed");
		pclipper->lpVtbl->SetHWnd(pclipper, 0, *D_API.pMainWnd);
		pDDSPrimary->lpVtbl->SetClipper(pDDSPrimary, pclipper);
		pclipper->lpVtbl->Release(pclipper);
	}
	else
	{
	    ZeroMemory(&ddscaps, sizeof(DDSCAPS2));
	    ddscaps.dwCaps=DDSCAPS_BACKBUFFER;
		hres=pDDSPrimary->lpVtbl->GetAttachedSurface(pDDSPrimary, &ddscaps, &pDDSBack);
	    if (hres!=DD_OK)
			I_Error("GetAttachedSurface Failed");
	}
    I_InitScreenSurface();
    I_SetGamma();
    I_InitOverlay();
	if (InWindow)
		pDD->lpVtbl->FlipToGDISurface(pDD);
}

//FUDGE add screenshot support
void I_ScreenShot(void)
{
}

void I_SetGamma(void)
{
    ManualGamma=*D_API.pGammaLevel;
}
